/*
 * This file is in the public domain.
 * 
 */

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <exec/io.h>
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/semaphores.h>
#include <clib/exec_protos.h>
#include <inline/exec.h>
#include "mem.h"

#ifndef NDEBUG
#define DebugWrite(x,y,z,w) printf(y,z,w)
#define allocerror(descr,size) \
   DebugWrite(NULL, "Memory allocation error (%s) with size (%ld)\n", descr, size);
#define deallocerror(descr,p) \
   DebugWrite(NULL, "Memory deallocation error (%s) address (%p)\n", descr, p);
#else
#define DebugWrite(x,y,z,w)
#define allocerror(descr,size)
#define deallocerror(descr,p)
#endif

extern struct ExecBase *SysBase;

void lockmem(void);
void unlockmem(void);

struct MemoryBlock
{
    struct MemoryBlock *next;
    size_t size;
    void *address;
};

struct MemoryList
{
    struct MemoryBlock *first;
    struct MemoryBlock *last;
    long size;
    long count;
    long opcount;
};

static struct MemoryList*      list = NULL;
static struct SignalSemaphore* lock = NULL;

void* allocmem(size_t size)
{
    struct MemoryBlock *newblock;
    size_t allocsize;

    initmem();
    lockmem();
    
    if (list == NULL) {
        list =
            (struct MemoryList*)
            AllocVec(sizeof(struct MemoryList), MEMF_ANY | MEMF_CLEAR);

        if (!list) {
            allocerror("list", sizeof(struct MemoryList));
            unlockmem();
            return 0;
        }

        list->first = NULL;
        list->last = NULL;
        list->size = 0;
        list->count = 0;
    }

    // Align to bytes of 4
    allocsize = (size + 3) & ~0x03UL;

    newblock =
        (struct MemoryBlock*)
        AllocVec(sizeof(struct MemoryBlock), MEMF_ANY | MEMF_CLEAR);

    if (!newblock) {
        allocerror("block", sizeof(struct MemoryBlock));
        unlockmem();
        return 0;
    }

    newblock->address =
        (struct MemoryBlock*)
        AllocVec(allocsize, MEMF_ANY | MEMF_CLEAR);

    if (!newblock->address) {
        FreeVec(newblock);
        allocerror("memory", allocsize);
        unlockmem();
        return 0;
    }

    newblock->size = allocsize;
    newblock->next = NULL;

    if(list->first == NULL) {
        list->first = newblock;
        list->last = newblock;
    } else {
        list->last->next = newblock;
        list->last = newblock;
    }

    list->size += allocsize;
    list->count++;

    list->opcount++;
    if (list->opcount % 25 == 0) {
        DebugWrite(
            NULL,
            "Memory usage: %ld bytes allocated in %ld blocks.\n",
            list->size,
            list->count
        );
    }

    //DebugWrite(NULL, "Memory allocated at address (%p)\n", newblock->address);

    unlockmem();
    
    return newblock->address;
}

void freemem(void* block)
{
    struct MemoryBlock *current, *last;

    lockmem();

    if (list == NULL) {
        deallocerror("list", (void*)0);
        unlockmem();
        return;
    }

    if (block == NULL) {
        deallocerror("memory", (void*)0);
        unlockmem();
        return;
    }

    last = NULL;
    current = list->first;
    while (current != NULL && current->address != block) {
        last = current;
        current = current->next;
    }

    if (current == NULL) {
        deallocerror("address not found", block);
        unlockmem();
        return;
    }

    list->size -= current->size;
    list->count--;

    if (list->first == current) {
        list->first = NULL;
        list->last = NULL;
    } else if (list->last == current) {
        last->next = current->next;
        list->last = last;
    } else {
        last->next = current->next;
    }

    FreeVec(current->address);
    FreeVec(current);

    list->opcount++;

    unlockmem();

    //DebugWrite(NULL, "Memory deallocated at address (%p)\n", block);
}

void* reallocmem(void *block, size_t size)
{   
    struct MemoryBlock *current, *last;
    void *newmem = NULL;

    if (list == NULL) {
        deallocerror("list", (void*)0);
        return 0;
    }

    if (block == NULL) {
        newmem = allocmem(size);
        return newmem;
    }

    last = NULL;
    current = list->first;
    while (current != NULL && current->address != block) {
        last = current;
        current = current->next;
    }

    if (current == NULL) {
        deallocerror("address not found", block);
        return 0;
    }
    
    if (size != 0) {
        newmem = allocmem(size);
        bcopy(current->address, newmem, current->size);
    }
    
    freemem(current->address);
    
    return newmem;
}

void freeall()
{
    struct MemoryBlock *current, *next;

    if (list == NULL) {
        return;
    }

    lockmem();

    current = list->first;
    while (current != NULL) {
        next = current->next;
        FreeVec(current->address);
        FreeVec(current);
        current = next;
    }

    FreeVec(list);
    list = NULL;
    
    unlockmem();

    if(lock) {
        FreeMem(lock, sizeof(struct SignalSemaphore));
        lock = NULL;
    }
}

void initmem(void)
{
    if (lock == NULL)
    {
        lock = (struct SignalSemaphore*)AllocMem(
            sizeof(struct SignalSemaphore),
            MEMF_ANY|MEMF_PUBLIC|MEMF_CLEAR);

        InitSemaphore(lock);
    }
}

void lockmem(void)
{
    if (lock != NULL)
        ObtainSemaphore(lock);
}

void unlockmem(void)
{
    if (lock != NULL)
        ReleaseSemaphore(lock);
}

char *strdup(const char *s1)
{
    char *s2;
    size_t len = strlen(s1);
    s2 = allocmem(++len);

    if(s2 == NULL)
    {
        return NULL;
    }

    memcpy(s2, s1, --len);
    return s2;
}
